#!/usr/bin/env python3
# encoding: utf-8
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Henrik tom Wörden <h.tomwoerden@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
This module allows to generate template tables from JSON schemas.
"""
import argparse
import re
import sys
from abc import ABC, abstractmethod
from argparse import RawTextHelpFormatter
from enum import Enum
from typing import Union

from openpyxl import Workbook
from openpyxl.workbook.child import INVALID_TITLE_REGEX


class ColumnType(Enum):
    SCALAR = 1
    LIST = 2
    FOREIGN = 3
    IGNORE = 3


class RowType(Enum):
    COL_TYPE = 1
    PATH = 2
    IGNORE = 3


class TableTemplateGenerator(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def generate(self, schema: dict, foreign_keys: dict):
        """ generates a sheet definition from a given JSON schema

        Parameters:
        -----------
        schema: dict, given JSON schema
        foreign_keys: dict, a configuration that defines which attributes shall be used to create
                      additional columns when a list of references exists. Keys of the dict are
                      elements of the path within the JSON where the key is needed. Suppose we want
                      to distinguis Persons that are referenced by Trainings, foreign_keys must at
                      least contain the following: {"Training": {"Person": [key1, key2]}}.
                      Values wihtin the dicts can be a list representing the keys or
                      a dict that allows to set further foreign keys at lower depth.
                      In latter case if foreign keys exist at that level (
                      e.g. in the above example there might be further levels below "Person".),
                      then the keys can be set using the special "__this__" key.
                      Example: {"Training": {"__this__": ["date"], "Person": [key1, key2]}}
                      Here, "date" is the sole key for Training.
        """
        pass

    def _generate_sheets_from_schema(self, schema: dict, foreign_keys: dict = None
                                     ) -> dict[str, dict[str, list]]:
        """ generates a sheet definition from a given JSON schema

        Parameters
        ----------
        schema: dict
            given JSON schema
        foreign_keys: dict
            a configuration that defines which attributes shall be used to create
            additional columns when a list of references exists. See foreign_keys
            argument of TableTemplateGenerator.generate.

        Returns
        -------
            dict
            The structure of the dict, lets call it ``sheets`` is as follows:
            ``sheets[sheetname][colname]= (COL_TYPE, description, [path])``
            I.e. the top level dict contains sheet names as keys and dicts as values.
            The inner dict has column names as keys and tuples as values. The tuples have the
            column type as the first element, the description of the corresponding property as
            second and the path as a list as last.

        """
        if not ("type" in schema or "anyOf" in schema):
            raise ValueError("Inappropriate JSON schema: The following part should contain the "
                             f"'type' key:\n{schema}\n")
        if foreign_keys is None:
            foreign_keys = {}
        # here, we treat the top level
        # sheets[sheetname][colname]= (COL_TYPE, [path])
        sheets: dict[str, dict[str, tuple[str, list]]] = {}
        if "properties" not in schema:
            raise ValueError("Inappropriate JSON schema: The following part should contain "
                             f"the 'properties' key:\n{schema}\n")
        for RTname in schema["properties"].keys():
            sheets[RTname] = self._treat_schema_element(schema["properties"][RTname],
                                                        sheets, [RTname], foreign_keys)
        return sheets

    def _get_foreign_keys(self, foreign_keys: dict, path: list) -> list:
        """ returns the foreign keys that are needed at the location to which path points """
        cpath = list(path)
        keys = foreign_keys
        selected_keys = None
        while cpath:
            if keys is None:
                raise ValueError(f"A foreign key definition is missing for path:"
                                 f"\n{path}\n{foreign_keys}")
            if cpath[0] not in keys:
                raise ValueError(f"A foreign key definition is missing for path: \n{path}\n{keys}")
            keys = keys[cpath[0]]
            if isinstance(keys, list):
                selected_keys, keys = keys, None
            else:
                selected_keys, keys = None, keys
            cpath = cpath[1:]
        if isinstance(keys, dict) and "__this__" in keys:
            selected_keys = keys["__this__"]

        if selected_keys is None:
            raise ValueError(f"A foreign key definition is missing for path:"
                             f"\n{path}\n{foreign_keys}")
        return selected_keys

    def _treat_schema_element(self, schema: dict, sheets: dict = None, path: list = None,
                              foreign_keys: dict = None, level_in_sheet_name: int = 1
                              ) -> dict[str, tuple[str, list]]:
        """ recursively transforms elements from the schema into column definitions """
        if not ("type" in schema or "enum" in schema or "oneOf" in schema or "anyOf" in schema):
            raise ValueError("Inappropriate JSON schema: The following part should contain the "
                             f"'type' key:\n{schema}\n")

        ctype = ColumnType.SCALAR

        # if it is an array, value defs are in 'items'
        if 'type' in schema and schema['type'] == 'array':
            if 'type' in schema['items'] and schema['items']['type'] == 'object' and len(path) > 1:  # list of references; special treatment
                # we add a new sheet
                sheets[path[-1]] = self._treat_schema_element(schema['items'], sheets, path,
                                                              foreign_keys,
                                                              len(path))
                for c in self._get_foreign_keys(foreign_keys, path[:-1]):
                    sheets[path[-1]].update({c: (ColumnType.FOREIGN, f"see sheet '{path[0]}'", path[:-1]+[c])})
                # columns are added to the new sheet, thus we do not return columns
                return {}
            else:
                # it is a list of primitive types -> semi colon separated list
                schema = schema['items']
                ctype = ColumnType.LIST

        if 'oneOf' in schema:
            for el in schema['oneOf']:
                if 'type' in el:
                    schema = el

        if "properties" in schema:
            # recurse for each property
            cols = {}
            for pname in schema["properties"].keys():
                cols.update(self._treat_schema_element(
                    schema["properties"][pname], sheets, path+[pname], foreign_keys,
                    level_in_sheet_name))
            return cols
        else:
            description = schema['description'] if 'description' in schema else None

            # those are the leaves
            if 'type' not in schema:
                if 'enum' in schema:
                    return {".".join(path[level_in_sheet_name:]): (ctype, description, path)}
                if 'anyOf' in schema:
                    for d in schema['anyOf']:
                        # currently the only case where this occurs is date formats
                        assert d['type'] == 'string'
                        assert d['format'] == 'date' or d['format'] == 'date-time'
                    return {".".join(path[level_in_sheet_name:]): (ctype, description, path)}
            elif schema["type"] in ['string', 'number', 'integer', 'boolean']:
                if 'format' in schema and schema['format'] == 'data-url':
                    return {}  # file; ignore for now
                return {".".join(path[level_in_sheet_name:]): (ctype, description, path)}
            else:
                raise ValueError("Inappropriate JSON schema: The following part should define an"
                                 f" object with properties or a primitive type:\n{schema}\n")
        raise RuntimeError("This should not be reached. Implementation error.")


class XLSXTemplateGenerator(TableTemplateGenerator):
    def __init__(self):
        pass

    def generate(self, schema: dict, foreign_keys: dict, filepath: str):
        """ generates a sheet definition from a given JSON schema

        Parameters:
        -----------
        schema: dict, given JSON schema
        foreign_keys: dict, a configuration that defines which attributes shall be used to create
                      additional columns when a list of references exists. See foreign_keys
                      argument of TableTemplateGenerator.generate.
        filepath: str, the XLSX file will be stored under this path.
        """
        sheets = self._generate_sheets_from_schema(schema, foreign_keys)
        wb = self._create_workbook_from_sheets_def(sheets)
        wb.save(filepath)

    @staticmethod
    def _get_max_path_length(sheetdef: dict) -> int:
        """ returns the length of the longest path contained in the sheet definition

        see TableTemplateGenerator._generate_sheets_from_schema for the structure of the sheets
        definition dict
        You need to pass the dict of a single sheet to this function.
        """
        return max([len(path) for _, _, path in sheetdef.values()])

    @staticmethod
    def _get_ordered_cols(sheetdef: dict) -> list:
        """
        creates a list with tuples (colname, column type, path) where the foreign keys are first
        """
        ordered_cols = []
        # first foreign cols
        for colname, (ct, desc, path) in sheetdef.items():
            if ct == ColumnType.FOREIGN:
                ordered_cols.append((colname, ct, desc, path))
        # now the other
        for colname, (ct, desc, path) in sheetdef.items():
            if ct != ColumnType.FOREIGN:
                ordered_cols.append((colname, ct, desc, path))

        return ordered_cols

    def _create_workbook_from_sheets_def(self, sheets: dict[str, dict[str, tuple[str, list]]]):
        wb = Workbook()
        assert wb.sheetnames == ["Sheet"]
        for sn, sheetdef in sheets.items():
            ws = wb.create_sheet(re.sub(INVALID_TITLE_REGEX, '_', sn))
            # first row will by the COL_TYPE row
            # first column will be the indicator row with values COL_TYPE, PATH, IGNORE
            # the COL_TYPE row will be followed by as many PATH rows as needed

            max_path_length = self._get_max_path_length(sheetdef)
            header_index = 2+max_path_length
            description_index = 3+max_path_length

            # create first column
            ws.cell(1, 1, RowType.COL_TYPE.name)
            for index in range(max_path_length):
                ws.cell(2+index, 1, RowType.PATH.name)
            ws.cell(header_index, 1, RowType.IGNORE.name)
            ws.cell(description_index, 1, RowType.IGNORE.name)

            ordered_cols = self._get_ordered_cols(sheetdef)

            # create other columns
            for index, (colname, ct, desc, path) in enumerate(ordered_cols):
                ws.cell(1, 2+index, ct.name)
                for pi, el in enumerate(path):
                    ws.cell(2+pi, 2+index, el)
                ws.cell(header_index, 2+index, colname)
                if desc:
                    ws.cell(description_index, 2+index, desc)

            # hide special rows
            for index, row in enumerate(ws.rows):
                if not (row[0].value is None or row[0].value == RowType.IGNORE.name):
                    ws.row_dimensions[index+1].hidden = True

            # hide special column
            ws.column_dimensions['A'].hidden = True

        # remove initial sheet
        del wb['Sheet']

        # order sheets
        # for index, sn in enumerate(sorted(wb.sheetnames)):
        # wb.move_sheet(sn, index-wb.index(wb[sn]))
        # reverse sheets
        for index, sn in enumerate(wb.sheetnames[::-1]):
            wb.move_sheet(sn, index-wb.index(wb[sn]))

        return wb


def parse_args():
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=RawTextHelpFormatter)
    parser.add_argument("path",
                        help="the subtree of files below the given path will "
                        "be considered. Use '/' for everything.")

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    sys.exit(main(args))
